/* The Computer Language Benchmarks Game
   http://shootout.alioth.debian.org/

   contributed by Greg Buchholz (original iterative version)
   modified by Yannick Gingras (made tail recursive)
   adjusted slightly for Utah CS 4960-01 by Matthew Flatt
*/

#include <stdio.h>
#include <stdlib.h>
#include <omp.h>

struct color { int r, g, b; } cols[]
= { { 255, 0, 0 },
    { 0, 255, 0 },
    { 0, 0, 255 },
    { 255, 255, 0 },
    { 255, 0, 255 },
    { 0, 255, 255 } };

int iterate(double Zr, double Zi,
	    double Tr, double Ti,
	    double Cr, double Ci,
	    int iter)
{

  if (iter == 0)
    return 1;

  Zi = 2.0 * Zr * Zi + Ci;
  Zr = Tr - Ti + Cr;
  Tr = Zr * Zr;
  Ti = Zi * Zi;

  if (Tr + Ti > 4.0)
    return 0;

  return iterate (Zr, Zi, Tr, Ti, Cr, Ci, iter - 1);
}

int main (int argc, char **argv)
{
  int w, h, buflen, rowlen;
  char *buf;
  int y, x;
  double wrat, hrat;

  w = h = atoi (argv[1]);
  wrat = 2.0 / w;
  hrat = 2.0 / h;
  rowlen = w * 3;
  buflen = rowlen * h;
  buf = (char *) malloc (buflen);

  /* The mandelbort image is written to buf, which is an array
     of rowlen * h bytes, where each rowlen bytes is one line
     of the image (representing w pixels). */

  printf ("P6\n%d %d\n255\n", w, h);

#pragma omp parallel for collapse(2) schedule(static, 10)
  for (y = 0; y < h; ++y)
    {
      for (x = 0; x < w; ++x)
        {
          double Ci = y * hrat - 1.0;
          char *pbuf = buf + (rowlen * y) + (x * 3);
          int id = omp_get_thread_num();
          double Cr = x * wrat - 1.5;
            
          if (iterate (0.0, 0.0, 0.0, 0.0, Cr, Ci, 50)) {
            *(pbuf++) = cols[id].r;
            *(pbuf++) = cols[id].g;
            *(pbuf++) = cols[id].b;
          } else {
            *(pbuf++) = cols[id].r / 2;
            *(pbuf++) = cols[id].g / 2;
            *(pbuf++) = cols[id].b / 2;
          }
        }

    }

  fwrite (buf, buflen, 1, stdout);
  return 0;
}
